#!/usr/bin/env python2
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See LICENSE for more details.
#
# Copyright: Red Hat Inc. 2017
# Author: Benjamin Berg <bberg@redhat.com>

import multiprocessing
import Queue
import subprocess
import threading
import sys
import pwd, os
import logging
import signal
import os.path
import time

from avocado.utils import process
from avocado.core import output
from . import ui

TEST_USER="fedora-laptop-testing"
GDM_CONFIG='/etc/gdm/custom.conf'

def _copy_fd_to_filelike_forever(_in_fd, _out):
    try:
        while True:
            data = os.read(_in_fd, 1024)
            if len(data) == 0:
                return

            _out.write(data)
    except EOFError:
        return


# So that it can be pickled for another layer of security that we clean up
# properly
def _restore_config():
    process.system('passwd -l fedora-laptop-testing')

    if not os.path.exists(GDM_CONFIG + '.pre-fedora-laptop-testing'):
        return

    os.unlink(GDM_CONFIG)
    os.rename(GDM_CONFIG + '.pre-fedora-laptop-testing', GDM_CONFIG)

class Session():
    def __init__(self, test, start_hidden=False):
        self.app_ctrl = None
        self._test = test
        self._start_hidden = start_hidden

        if 'NO_SESSION' in os.environ and os.environ['NO_SESSION'] in ['1', 'y', 'yes', 'true']:
            self.no_session = True
        else:
            self.no_session = False

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, type, value, traceback):
        self.stop()

    def _gdm_restart(self):
        # Try really hard to restart GDM because of the issues raised in
        # https://bugzilla.gnome.org/show_bug.cgi?id=784800
        try:
             process.system('systemctl stop gdm')
        except process.CmdError:
             pass
        try:
            process.system('loginctl terminate-user gdm')
        except process.CmdError:
             pass
        try:
            process.system('systemctl start gdm')
        except process.CmdError:
             pass

    def start(self):
        if not self.no_session:
            # Start a new session so that we can guarantee our inhibitor is
            # actually active (otherwise it might get cleaned up by avocado)
            try:
                process.system('loginctl unlock-sessions')
                process.system('loginctl terminate-user %s' % TEST_USER)
            except process.CmdError:
                pass

            # XXX: This sleep prevent the new session from being locked if the old
            #      one was locked.
            time.sleep(1)

            self._modify_config()

            self._gdm_restart()

        tries = 10
        self.app_ctrl = None
        while self.app_ctrl is None:
            time.sleep(1)
            tries -= 1
            try:
                self.app_ctrl = ui.AppCtrl(self._test, self, self._start_hidden)
                self.app_ctrl.wait_started()

            except (ui.UIError, subprocess.CalledProcessError) as e:
                if self.app_ctrl is not None:
                    try:
                        self.app_ctrl.stop()
                    finally:
                        self.app_ctrl = None

                if tries <= 0:
                    import traceback
                    traceback.print_exc()

                    try:
                        self.stop()
                    except:
                        print('Error while stopping the session after failing to start it (this is kind of expected)')
                        traceback.print_exc()
                        pass

                    raise e

        self._test.log.info('Session and main GUI initialized')

    def stop(self):
        if not self.no_session:
            self._restore_config()

        if self.app_ctrl is not None:
            self.app_ctrl.stop()

        if not self.no_session:
            try:
                process.system('loginctl terminate-user %s' % TEST_USER)
            except process.CmdError:
                pass
            self._gdm_restart()

    def _modify_config(self):
        self._modify_gdm_config()

        process.system('passwd -d fedora-laptop-testing')

    def _modify_gdm_config(self):
        # Just in case
        self._restore_config()

        self._test.runner_queue.put({
            "func_at_exit": _restore_config,
            "once": True})

        os.rename(GDM_CONFIG, GDM_CONFIG + '.pre-fedora-laptop-testing')
        f = open(GDM_CONFIG, 'w')
        f.write('''
[daemon]
AutomaticLoginEnable=True
AutomaticLogin=%s

[security]
[xdmcp]
[chooser]
[debug]
''' % TEST_USER)

    def _restore_config(self):
        _restore_config()

    def _setup_env(self, func, stdout, stderr, *args, **kwargs):
        """
        Wrapper to setup the global scope for the remote process.
        Needs to:
         1. Steal environment variables from session
            (XDG_RUNTIME_DIR, WAYLAND_DISPLAY, ...)
         2. Switch UID to the target user
         3. Redirect stdout/stderr (TODO: and other logging facilities)
         4. Run the target function
        """

        del os.environ['GDK_BACKEND']

        os.dup2(stdout[1], 1)
        os.dup2(stderr[1], 2)

        os.close(stdout[1])
        os.close(stderr[1])
        os.close(stdout[0])
        os.close(stderr[0])

        try:
            if not self.no_session:
                uid = pwd.getpwnam(TEST_USER).pw_uid
                gid = pwd.getpwnam(TEST_USER).pw_gid

                # Grab the sessions environment to steal it later
                session_pid = process.system_output("pgrep -n -U %i gnome-session" %uid)
                session_pid = int(session_pid.strip())
                environ = open('/proc/%i/environ' % session_pid).read()

                # TODO: Disable logging to files for all log handlers (or ensure all
                #       logfiles are opened before setuid is run)!

                os.setgid(gid)
                os.setuid(uid)

                for envvar in environ.split('\0'):
                    var = envvar.split('=', 1)
                    if len(var) == 2:
                        os.environ[var[0]] = var[1]

            # Disable avocado signal handlers and ensure that we do not exit
            # cleanly.
            def term_handler(*args):
                sys.stderr.write("Got SIGTERM in session process")
                os._exit(1)

            signal.signal(signal.SIGTERM, term_handler)

            res = func(*args, **kwargs)
        except:
            import traceback
            info = traceback.format_exc()

            sys.stderr.write(info)

            os._exit(1)

        os._exit(res)

    def spawn_process(self, func, *args, **kwargs):
        stdout = os.pipe()
        stderr = os.pipe()

        nargs = [func, stdout, stderr]
        nargs.extend(args)

        proc = multiprocessing.Process(target=self._setup_env,
                                       args=nargs, kwargs=kwargs)
        proc.start()
        self._test.log.info("Spawned session subprocess with PID %i", proc.pid)

        threading.Thread(target=_copy_fd_to_filelike_forever, args=(stdout[0], sys.stdout)).start()
        threading.Thread(target=_copy_fd_to_filelike_forever, args=(stderr[0], sys.stderr)).start()

        os.close(stdout[1])
        os.close(stderr[1])

        return proc

    def spawn_subprocess(self, func, *args, **kwargs):
        # This should be rewritten to be nicer!
        stdout = os.pipe()
        stderr = (os.dup(stdout[0]), os.dup(stdout[1]))

        nargs = [func, stdout, stderr]
        nargs.extend(args)

        proc = multiprocessing.Process(target=self._setup_env,
                                       args=nargs, kwargs=kwargs)
        proc.start()
        self._test.log.info("Spawned session subprocess with PID %i", proc.pid)

        os.close(stdout[1])
        os.close(stderr[1])

        return proc, stdout[0]

    def _convert_value(self, value):
        if isinstance(value, bool):
            return 'true' if value else 'false'
        elif isinstance(value, list):
            return '[' + ','.join(self._convert_value(v) for v in value) + ']'
        elif isinstance(value, tuple):
            return '(' + ','.join(self._convert_value(v) for v in value) + ')'
        elif isinstance(value, int):
            return str(value)
        else:
            return "'" + value + "'"

    def set_dconf_key(self, key, value):
        self.set_dconf_key_raw(key, self._convert_value(value))

    def set_dconf_key_raw(self, key, value):
        proc = self.spawn_process(
                lambda : os.execl('/usr/bin/dconf', 'dconf', 'write', key, value))
        proc.join()

    def get_dconf_key_raw(self, key):
        proc, output = self.spawn_subprocess(
                lambda : os.execl('/usr/bin/dconf', 'dconf', 'read', key))

        result = os.read(output, 10000)
        proc.join()
        return result

