# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See LICENSE for more details.
#
# Copyright: Red Hat Inc. 2017
# Author: Benjamin Berg <bberg@redhat.com>

import time

from gi.repository import Gtk, Gdk, GLib, Gio
from .structs import *

class DBusSignal(Signal):
    def __init__(self, bus_name, obj, interface, method, params):
        self.bus_name = bus_name
        self.obj = obj
        self.interface = interface
        self.method = method
        self.params = params

    def __getstate__(self):
        res = dict(self.__dict__)
        if res['params'] is not None:
            res['params'] = res['params'].get_type_string(), res['params'].get_data_as_bytes().get_data()

        return res

    def __setstate__(self, state):
        for k, v in state.iteritems():
            if k == 'params' and v is not None:
                v = GLib.Variant.new_from_bytes(GLib.VariantType.new(v[0]), GLib.Bytes.new(v[1]), True)

            setattr(self, k, v)


class DBusCancelSignal(Signal):
    pass

class DBusReplyEvent(Event):
    def __init__(self, reply=None, error=False):
        self.reply = reply
        self.error = error

    def __getstate__(self):
        res = dict(self.__dict__)
        if res['reply'] is not None:
            res['reply'] = res['reply'].get_type_string(), res['reply'].get_data_as_bytes().get_data()

        return res

    def __setstate__(self, state):
        for k, v in state.iteritems():
            if k == 'reply' and v is not None:
                v = GLib.Variant.new_from_bytes(GLib.VariantType.new(v[0]), GLib.Bytes.new(v[1]), True)

            setattr(self, k, v)

class DBusUI(TestUI):

    __metaclass__ = UIRegister
    ui_name = "dbus"

    def __init__(self, app):
        self.app = app
        self.session_bus = None


    def ensure_bus(self):
        if self.session_bus is None:
            self.session_bus = Gio.bus_get_sync(Gio.BusType.SESSION)

    def call_finished(self, bus, result):
        try:
            reply = self.session_bus.call_finish(result)

            self.app.send(DBusReplyEvent(reply=reply))
        except Exception as e:
            error = str(e)
            self.app.send(DBusReplyEvent(error=error))

    def recv(self, signal):
        if isinstance(signal, DBusCancelSignal):
            self.cancel.cancel()
            self.cancel = None

        if not isinstance(signal, DBusSignal):
            return

        self.ensure_bus()
        self.cancel = Gio.Cancellable()

        try:
            self.session_bus.call(
                signal.bus_name, signal.obj,
                signal.interface, signal.method, signal.params,
                None, # reply type
                Gio.DBusCallFlags.NONE,
                5000, # Timeout in ms
                self.cancel,
                self.call_finished)
        except:
            import traceback
            error = traceback.format_exc()

            self.app.send(DBusReplyEvent(error=error))


class DBusCtrl(TestUICtrl):

    __metaclass__ = UICtrlRegister
    ui_name = "dbus"

    def __init__(self, app_ctrl):
        self.app_ctrl = app_ctrl
        self.test = app_ctrl.test
        self.log = self.test.log

    def recv(self, event):
        if not isinstance(event, DBusReplyEvent):
            return

        if event.error:
            raise UIError('Error doing DBus call:' + str(event.error))

    def wait_reply(self, timeout=1, **kwargs):
        return self.app_ctrl.wait_event(timeout, lambda event: isinstance(event, DBusReplyEvent))

    def call(self, bus_name, obj, interface, method, params):
        self.app_ctrl.send(DBusSignal(bus_name, obj, interface, method, params))

    def call_wait(self, bus_name, obj, interface, method, params, timeout=1, retry=0):
        self.app_ctrl.read_pending()

        while True:
            self.call(bus_name, obj, interface, method, params)
            try:
                return self.wait_reply(timeout=timeout)
            except TimeoutError as e:
                self.app_ctrl.send(DBusCancelSignal())
                raise e
            except:
                if retry == 0:
                    raise
                retry -= 1
                time.sleep(1)

