# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See LICENSE for more details.
#
# Copyright: Red Hat Inc. 2017
# Author: Benjamin Berg <bberg@redhat.com>

import time

from gi.repository import Gtk, Gdk, GLib, Gio
from .structs import *

class DialogSignal(Signal):
    def __init__(self, title, markup, icon, buttons, default_response=None, include_entry=False):
        self.title = title
        self.markup = markup
        self.icon = icon
        btns = []
        for btn in buttons:
            if isinstance(btn, Gtk.ResponseType):
                btns.append(int(btn))
            else:
                btns.append(btn)
        self.buttons = btns
        self.default_response = int(default_response)
        self.include_entry = include_entry

class DialogReplyEvent(Event):
    def __init__(self, response):
        self.response = int(response)
        self.text = False

class PromptUI(TestUI):

    __metaclass__ = UIRegister
    ui_name = "prompt"

    def __init__(self, app):
        self.app = app

    def recv(self, signal):
        if not isinstance(signal, DialogSignal):
            return

        flags = Gtk.DialogFlags.MODAL | Gtk.DialogFlags.DESTROY_WITH_PARENT
        parent = self.app.main_win


        if signal.markup is None:
            label = signal.title
        else:
            label = signal.markup

        dialog = Gtk.MessageDialog(title=signal.title, parent=parent, flags=flags,
                                   buttons=signal.buttons, text=label, use_markup=True)
        if signal.default_response is not None:
            dialog.set_default_response(signal.default_response)

        if signal.icon:
            image = Gtk.Image.new_from_icon_name(signal.icon, Gtk.IconSize.DIALOG)

        content = dialog.get_content_area()

        if signal.include_entry:
            entry = Gtk.Entry()
            entry.set_activates_default(True)
            content.pack_end(entry, False, True, 5)
            entry.show()

        response = dialog.run()

        event = DialogReplyEvent(response)
        if signal.include_entry:
            event.text = entry.get_text()

        dialog.destroy()

        self.app.send(event)

class PromptCtrl(TestUICtrl):

    __metaclass__ = UICtrlRegister
    ui_name = "prompt"

    def __init__(self, app_ctrl):
        self.app_ctrl = app_ctrl
        self.test = app_ctrl.test
        self.log = self.test.log

    def recv(self, event):
        if not isinstance(event, DialogSignal):
            return

    def wait_reply(self, timeout=-1, **kwargs):
        event = self.app_ctrl.wait_event(timeout, lambda event: isinstance(event, DialogReplyEvent))
        return event

    def prompt(self, title, markup, icon, buttons, default_response=None, include_entry=False):
        self.app_ctrl.send(DialogSignal(title, markup, icon, buttons, default_response, include_entry))

    def prompt_wait(self, title, markup, icon, buttons, default_response=None, include_entry=False):
        self.app_ctrl.read_pending()

        self.prompt(title, markup, icon, buttons, default_response, include_entry)

        return self.wait_reply()

    def wait(self, title=None, markup=None, icon=None):
        if markup is None and title is None:
            markup = 'The next step requires your attention, please confirm to continue.'
        if title is None:
            title = 'Waiting for user interaction'
        self.prompt_wait(title, markup, icon, buttons=("Continue", Gtk.ResponseType.ACCEPT), default_response=Gtk.ResponseType.ACCEPT)

    def ok_or_skip(self, title=None, markup=None, icon=None):
        """Skip test if the user says so.

        Note that cleanup will be done automatically by this function."""
        if title is None:
            title = "Run this test?"
        if markup is None:
            markup = title

        event = self.prompt_wait(title, markup, icon, buttons=("Test not applicable", Gtk.ResponseType.REJECT, "Skip", Gtk.ResponseType.CANCEL, "Run Test", Gtk.ResponseType.ACCEPT), default_response=Gtk.ResponseType.ACCEPT)
        if event.response != Gtk.ResponseType.ACCEPT:
            # Ensure cleanup
            self.test.tearDown()

            if event.response == Gtk.ResponseType.REJECT:
                self.test.skip('User marked test as not applicable')

            # CANCEL, i.e. Esc key or "Skip"
            self.test.skip('User chose to skip test')

    def ok_or_fail(self, title=None, markup=None, extra=None, icon=None):
        if markup is None and title is None:
            markup = 'Was the test successfull or not?\n' + \
                     'Please enter additional information if the test failed. ' + \
                     'If unsure enter a comment and mark the test as failure.'
            if extra:
                markup += '\n\n'
                markup += extra
        if title is None:
            title = 'Test Result'

        event = self.prompt_wait(title, markup, icon, buttons=("Failure", Gtk.ResponseType.NO, "Success", Gtk.ResponseType.YES), default_response=Gtk.ResponseType.YES, include_entry=True)
        if event.response != Gtk.ResponseType.YES:
            self.test.fail('Marked as failure by user: ' + str(event.text))
        elif event.text:
            self.test.log.warning('User entered text: ' + str(event.text))

