# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See LICENSE for more details.
#
# Copyright: Red Hat Inc. 2017
# Author: Benjamin Berg <bberg@redhat.com>

from __future__ import division

import gi
gi.require_version('Gtk', '3.0')
gi.require_version('Gst', '1.0')
from gi.repository import Gst, Gtk, GLib, GObject

import struct
import numpy as np
import os

_gst_initialized = False
def _ensure_gst_init():
    global _gst_initialized
    if _gst_initialized:
        return
    Gst.init([])
    _gst_initialized = True

    #Gst.debug_set_active(True)
    #Gst.debug_set_default_threshold(Gst.DebugLevel.WARNING)



class PlayRecordAudio(GObject.Object):
    # Detect native endianness, will also work if wrong
    if struct.pack('!h', 0x0102) == b'\x01\x02':
        fmt = 'S16LE'
        npfmt = '<i2'
    else:
        fmt = 'S16BE'
        npfmt = '>i2'

    def __init__(self, sample_freq=44100, samples=44100):
        GObject.Object.__init__(self)
        _ensure_gst_init()

        self.sample_freq = sample_freq
        self._create_pipeline()
        self.resultsink.connect('handoff', self._result_handoff_cb)

        self._received_samples = 0
        self._samples = samples
        self._result_buffers = []
        self._np_samples = None
        self._playing = False
        self._stop_handler_id = 0

    def play(self):
        assert not self._playing
        #assert self.pipeline.get_state(0)[1] in [ Gst.State.NULL, Gst.State.PAUSED ]
        self._received_samples = 0
        self._result_buffers = []
        self.pipeline.set_state(Gst.State.PLAYING)
        self.pipeline.sync_children_states()
        self._playing = True

    def _create_pipeline(self):

        self.pipeline = Gst.Pipeline.new('test')

        testsource = self._get_source_pipeline()
        micsource = Gst.ElementFactory.make('pulsesrc', 'pulsesrc')

        self.pipeline.add(testsource)
        self.pipeline.add(micsource)

        testsrctee = Gst.ElementFactory.make('tee', 'testsrctee')
        self.pipeline.add(testsrctee)


        interleave = Gst.ElementFactory.make('interleave', 'interleave')
        self.pipeline.add(interleave)


        self.resultsink = Gst.ElementFactory.make('fakesink', 'resultsink')
        self.resultsink.set_property('signal-handoffs', True)

        pulsesink = Gst.ElementFactory.make('pulsesink', 'pulsesink')
        pulsesink.set_property('slave-method', 'none')
        pulsesink.set_property('async', True)
        self.pipeline.add(pulsesink)


        self.pipeline.add(self.resultsink)


        queue1 = Gst.ElementFactory.make('queue')
        queue2 = Gst.ElementFactory.make('queue')
        self.pipeline.add(queue1)
        self.pipeline.add(queue2)


        mono_caps = Gst.Caps.from_string ('audio/x-raw,channels=1,rate=%i,layout=interleaved,format=%s' % (self.sample_freq, self.fmt))
        filter1 = Gst.ElementFactory.make ('capsfilter', 'filter1')
        filter2 = Gst.ElementFactory.make ('capsfilter', 'filter2')
        filter1.set_property('caps', mono_caps)
        filter2.set_property('caps', mono_caps)
        self.pipeline.add(filter1)
        self.pipeline.add(filter2)

        # And link everything.
        #   testsrc -> testsrctee -> pulsesink
        #                        \-> queue -> capsfilter \
        #   pulsesrc ->              queue -> capsfilter +-> interleave -> resultsink
        testsource.link(testsrctee)
        testsrctee.link(pulsesink)
        testsrctee.link(queue1)

        micsource.link(queue2)

        queue1.link(filter1)
        queue2.link(filter2)

        filter1.link(interleave)
        filter2.link(interleave)

        interleave.link(self.resultsink)

    def _get_source_pipeline(self):
        raise NotImplemented

    def _result_handoff_cb(self, sink, buf, pad):
        if not self._playing:
            return

        self._received_samples += buf.get_size() // 4

        result, mapping = buf.map(Gst.MapFlags.READ)
        self._result_buffers.append(mapping.data)

        if self._received_samples >= self._samples and not self._stop_handler_id:
            self._stop_handler_id = GLib.idle_add(self._stop_pipeline)

    def _stop_pipeline(self):
        assert self._playing
        self.pipeline.set_state(Gst.State.NULL)
        self.pipeline.sync_children_states()
        self._playing = False
        self.done.emit()

        self._stop_handler_id = 0
        return False

    @GObject.Signal
    def done(self):
        self._np_samples = self._get_np_samples()
        self._done()

    def _done(self):
        # Handler for doing additional tasks in subclasses
        pass

    def _get_np_samples(self):
        # Assume this is executed after the "done" signal has been emited
        result = np.empty((2, self._samples), dtype='int16', order='F')

        offset = 0
        for buf in self._result_buffers:
            data = np.fromstring(buf, dtype=self.npfmt)

            data = data.reshape((2, data.shape[0] // 2), order='F')

            copy_length = data.shape[1]
            if offset + copy_length > result.shape[1]:
                copy_length = result.shape[1] - offset

            result[:,offset:offset+copy_length] = data[:,:copy_length]

            offset += copy_length

        assert offset == self._samples

        # Convert to floating point
        result = result / (2**15)
        return result

    @property
    def result(self):
        return self._np_samples

class SineAmplitudeTest(PlayRecordAudio):
    def __init__(self, sample_freq=44100, samples=44100):
        # This should be around 430Hz
        self.freq = sample_freq / 1024 * 10

        PlayRecordAudio.__init__(self, sample_freq, samples)

    def _get_source_pipeline(self):
        self._source = Gst.ElementFactory.make ('audiotestsrc', 'testsrc')
        self._source.set_property('wave', 'sine')
        self._source.set_property('freq', self.freq)
        self._source.set_property('volume', 1.0)
        return self._source

    def set_volume(self, volume):
        testsrc = self.pipeline.get_by_name('testsrc')
        testsrc.set_property('volume', volume)

    def _done(self):
        result = self.result

        original = result[0]
        recording = result[1]

        samples = len(original)
        # Use the end of the FFT rather than the start
        # (as no input might be at the start)
        self._fft_len = int(samples * 0.8) // 1024 * 1024
        offset = (samples - self._fft_len)
        self._fft = np.fft.rfft(result[:,offset:offset + self._fft_len])

    def get_gain(self):
        test_freq_bin = int(self.freq / (self.sample_freq / self._fft_len))

        magnitude = np.abs(self._fft)
        magnitude = magnitude[:,test_freq_bin]

        gain = magnitude[1] / magnitude[0]

        return gain

    def get_THD_F(self, max_harmonic=999):
        test_freq_bin = int(self.freq / (self.sample_freq / self._fft_len))

        # We only need this for the input signal
        magnitude = np.abs(self._fft[1])
        reference_magnitude = magnitude[test_freq_bin]

        blub = int(2*self.freq / (self.sample_freq / self._fft_len))
        blub = int(3*self.freq / (self.sample_freq / self._fft_len))

        harmonics = magnitude[2*test_freq_bin:max_harmonic*test_freq_bin+1:test_freq_bin]

        # This could probably be optimised a bit, but that doesn't matter here
        return np.sqrt(np.sum(np.square(harmonics))) / reference_magnitude

def set_input_mute(mute):
    os.system(r'pacmd set-source-mute $( pacmd info | sed -n "s/Default source name:\s*\(.*\)/\1/p" ) %i' % int(mute))

def set_input_volume(volume):
    volume = int(65535 * volume)
    os.system(r'pacmd set-source-volume $( pacmd info | sed -n "s/Default source name:\s*\(.*\)/\1/p" ) %i' % volume)

def set_output_mute(mute):
    os.system(r'pacmd set-sink-mute $( pacmd info | sed -n "s/Default sink name:\s*\(.*\)/\1/p" ) %i' % int(mute))

def set_output_volume(volume):
    volume = int(65535 * volume)
    os.system(r'pacmd set-sink-volume $( pacmd info | sed -n "s/Default sink name:\s*\(.*\)/\1/p" ) %i' % volume)

if __name__ == '__main__':
    test = SineAmplitudeTest(samples=40000)

    set_input_mute(False)
    set_output_mute(False)

    test_params = [
#            ( 1.0, 0.01, 1.0 ),
#            ( 1.0, 0.2, 1.0 ),
#            ( 1.0, 0.4, 1.0 ),
#            ( 1.0, 0.6, 1.0 ),
#            ( 0.6 / 0.8, 0.8, 1.0 ),
#            ( 0.6 / 1.0, 1.0, 1.0 ),
            ( 1.0, 0.6, 0.8 ),
            ( 1.0, 0.6, 0.6 ),
            ( 1.0, 0.6, 0.4 ),
            ( 1.0, 0.6, 0.2 ),
            ( 1.0, 0.6, 0.1 ),
            ( 1.0, 0.6, 0.05 ),
            ( 1.0, 0.6, 0.025 ),
        ]
    test_idx = 0

    def set_test_params(params):
        test.set_volume(params[0]**3)
        set_output_volume(params[1])
        set_input_volume(params[2])

    def done_cb(obj):
        global test_idx
        gain = obj.get_gain()

        if gain > 0:
            gain_db = 20 * np.log10(gain) - 20*np.log10(test_params[test_idx][1]**3) - 20*np.log10(test_params[test_idx][2]**3)
        else:
            gain_db = - np.inf
        print('Signal: %2.1f, Output volume: %2.1f, Input Volume: %2.1f' % (test_params[test_idx]))
        print('Gain: %2.1f dB\tTHD_F (up to 10*freq): %.1f %%' % (gain_db, obj.get_THD_F(10) * 100))

        test_idx += 1
        if test_idx < len(test_params):
            set_test_params(test_params[test_idx])
            test.play()
        else:
            Gtk.main_quit()

    set_test_params(test_params[test_idx])

    test.connect('done', done_cb)
    test.play()


    Gtk.main()

    print('')
