#!/usr/bin/env python2

from fed_laptoptest import edid

import os
import sys

import gi
gi.require_version('GUdev', '1.0')
from gi.repository import GUdev, GObject, GLib

class Card(object):
    def __init__(self, drm_monitor, udev):
        self._drm_monitor = drm_monitor
        self.udev = udev
        self._outputs = dict()

    @property
    def descr(self):
        parent = self.udev.get_parent()
        model_name = parent.get_property('ID_MODEL_FROM_DATABASE')
        vendor_name = parent.get_property('ID_VENDOR_FROM_DATABASE')

        if model_name and vendor_name:
            return '{:s} - {:s}'.format(vendor_name, model_name)
        elif vendor_name:
            return '{:s} ({:s})'.format(vendor_name, self.udev.get_name())
        else:
            return self.udev.get_name()

    @property
    def name(self):
        return self.udev.get_name()

    @property
    def driver(self):
        return self.udev.get_parent().get_sysfs_attr('driver')

    def _sync(self):
        for output in self._outputs.itervalues():
            output._sync()

class Output(object):
    def __init__(self, drm_monitor, card, udev):
        self._drm_monitor = drm_monitor
        self.card = card
        self.udev = udev
        self.card._outputs[self.name] = self

        self._status = None
        self._edid = None

        self._sync()

    @property
    def connected(self):
        return self._status == 'connected'

    @property
    def name(self):
        return self.udev.get_name().split('-', 1)[1]

    @property
    def monitor(self):
        if self._edid is None:
            return None

        try:
            edid_obj = edid.Edid(self._edid)
            return edid_obj.model
        except ValueError:
            print('Invalid EDID information on %s'.format(self.name))
            return None

    def _resync(self):
        self._sync()
        return GLib.SOURCE_REMOVE

    def _sync(self):
        # GUDev caches the results and there is no uevent for the subdevice
        self.udev = self._drm_monitor._client.query_by_sysfs_path(self.udev.get_sysfs_path())
    
        old_status = self._status
        old_edid = self._edid

        self._status = self.udev.get_sysfs_attr('status')
        try:
            self._edid = open(os.path.join(self.udev.get_sysfs_path(), 'edid')).read()
        except OSError:
            self._edid = None
            sys.stderr.write('Error reading EDID information on %s, will retry again later.'.format(self.name))
            GLib.timeout_add(500, self._resync)

        if old_status != self._status or old_edid != self._edid:
            self._drm_monitor.output_changed.emit(self)

class DRMMonitor(GObject.Object):
    def __init__(self):
        GObject.Object.__init__(self)

        self._cards = dict()

        self._client = GUdev.Client.new(['drm'])
        self._client.connect('uevent', self._uevent)

        # Initialize the list from an idle handler so that any connect'ed
        # signals will be called!
        GLib.idle_add(self._idle_init, priority=GLib.PRIORITY_HIGH)

    def _idle_init(self):
        for udev in self._client.query_by_subsystem('drm'):
            parent = udev.get_parent()
            if udev.get_devtype() == 'drm_minor':
                card = Card(self, udev)
                self._cards[card.name] = card
            elif parent and self._get_card_by_name(parent.get_name()) is not None:
                card = self._get_card_by_name(udev.get_parent().get_name())
                # Assume it is an output, will add itself to the card
                output = Output(self, card, udev)
            elif parent is None:
                print('Parent of {:s} is None, ignoring'.format(udev.get_name()))

        return GLib.SOURCE_REMOVE


    def _sync(self):
        for c in self._cards.itervalues():
            c._sync()


    def _get_card_by_name(self, name):
        try:
            return self._cards[name]
        except KeyError:
            return None

    def _uevent(self, client, action, device):
        c = self._get_card_by_name(device.get_name())
        if c is not None:
            c._sync()

    @GObject.Signal(arg_types=[object])
    def output_changed(self, output):
        pass


if __name__ == '__main__':
    gi.require_version('Gtk', '3.0')
    from gi.repository import Gtk

    monitor = DRMMonitor()
    def output_changed_cb(m, output):
        print(output.card.descr, output.card.driver, output.name, output.monitor, output.connected)
    monitor.connect('output-changed', output_changed_cb)
    Gtk.main()

