#!/usr/bin/env python2

from __future__ import division

from evdev import InputDevice, categorize, ecodes, list_devices
import argparse
import sys
from collections import namedtuple

class TouchpadRangeRes(namedtuple('TouchpadRangeBase', ['min', 'max', 'resolution'])):
    @property
    def size(self):
        if self.resolution > 0:
            return (self.max - self.min) / self.resolution
        else:
            return -1

class TouchpadRangeMM(namedtuple('TouchpadRangeBase', ['min', 'max', 'size'])):
    @property
    def resolution(self):
        if self.size <= 1:
            return -1
        if self.max - self.min <= 1:
            return -1

        return int((self.max - self.min) // self.size)

class Touchpad(InputDevice):

    def __init__(self, dev):
        InputDevice.__init__(self, dev)
        assert self.is_touchpad(self)

        self.cap_abs = { c[0] : c[1] for c in self.capabilities()[ecodes.EV_ABS] }

        self.multitouch = ecodes.ABS_MT_POSITION_X in self.cap_abs and ecodes.ABS_MT_POSITION_Y in self.cap_abs

        self.kernel_x = TouchpadRangeRes(self.cap_abs[ecodes.ABS_X].min, self.cap_abs[ecodes.ABS_X].max, self.cap_abs[ecodes.ABS_X].resolution)
        self.kernel_y = TouchpadRangeRes(self.cap_abs[ecodes.ABS_Y].min, self.cap_abs[ecodes.ABS_Y].max, self.cap_abs[ecodes.ABS_Y].resolution)

        self.measured_x = TouchpadRangeMM(999999, -999999, 0)
        self.measured_y = TouchpadRangeMM(999999, -999999, 0)

    def __str__(self):
        return 'Touchpad({:.2f}x{:.2f}mm{:s}, name={:s}, device={:s})'.format(self.kernel_x.size, self.kernel_y.size, ', multitouch' if self.multitouch else '', self.name, self.fn)

    @classmethod
    def is_touchpad(cls, evdev):
        cap = evdev.capabilities()
        if not ecodes.EV_ABS in cap:
            return False

        cap_abs = { c[0] : c[1] for c in cap[ecodes.EV_ABS] }

        if not ecodes.ABS_X in cap_abs or not ecodes.ABS_Y in cap_abs:
            return False

        return True

    @classmethod
    def find_touchpads(cls):
        touchpads = []
        for dev in list_devices():
            try:
                d = cls(dev)
                touchpads.append(d)
            except OSError:
                # Suppose a device that cannot be opened is not a touchpad
                # This happens e.g. on the X1 Yoga on RHEL 7.4 with the accelerometer
                pass
            except AssertionError:
                pass
        return touchpads

    def _update_bounds(self, event):
        if event.type != ecodes.EV_ABS:
            return

        if event.code in [ecodes.ABS_X, ecodes.ABS_MT_POSITION_X]:
            _min = min(self.measured_x.min, event.value)
            _max = max(self.measured_x.max, event.value)
            self.measured_x = TouchpadRangeMM(_min, _max, self.measured_x.size)

        if event.code in [ecodes.ABS_Y, ecodes.ABS_MT_POSITION_Y]:
            _min = min(self.measured_y.min, event.value)
            _max = max(self.measured_y.max, event.value)
            self.measured_y = TouchpadRangeMM(_min, _max, self.measured_y.size)

    def set_size(self, width, height):
        _min = self.measured_x.min
        _max = self.measured_x.max
        self.measured_x = TouchpadRangeMM(_min, _max, width)

        _min = self.measured_y.min
        _max = self.measured_y.max
        self.measured_y = TouchpadRangeMM(_min, _max, height)

    def read(self):
        for event in InputDevice.read(self):
            self._update_bounds(event)
            yield event

    def read_one(self):
        event = InputDevice.read_one(self)
        self._update_bounds(event)
        return event

def main():
    for touchpad in Touchpad.find_touchpads():
        print(touchpad)
        import cPickle
        cPickle.pickle(touchpad)

if __name__ == '__main__':
    main()

