#!/usr/bin/env python2
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See LICENSE for more details.
#
# Copyright: Red Hat Inc. 2017
# Author: Benjamin Berg <bberg@redhat.com>

import os
import os.path

from avocado import main

from fed_laptoptest.test import SessionTest, batch

import subprocess
import json

class FWTSBatch(SessionTest):
    """
    .. title:: Firmware Test Suite (batch)

    This test runs the FWTS firmware test suite in batch mode. No interaction
    will be required, please stand by until it finishes.

    1. .. class:: action-test

       Test is running, please stand by.

    :avocado: enable
    :avocado: tags=automatic
    :categories: firmware
    """

    def setUp(self):
        if not os.path.exists('/usr/bin/fwts'):
            self.skip('FWTS is not installed, skipping test.')
        SessionTest.setUp(self)
        self.fwts = None

    def tearDown(self):
        if self.fwts is not None:
            self.fwts.kill()
        SessionTest.tearDown(self)

    @batch
    def test(self):
        self.app_ctrl.status.mark_action('test')

        path = os.path.join(self.outputdir, 'fwts.json')
        self.fwts = subprocess.Popen(['/usr/bin/fwts', '-b', '-D',
                                          '--log-type', 'json',
                                          '-r', path],
                                     stdout=subprocess.PIPE)

        while self.fwts.returncode is None:
            line = self.fwts.stdout.readline()
            self.app_ctrl.read_pending()
            # Ignore non-start of block messages
            if not line:
                break

            if line != 'XXX\n':
                continue

            try:
                line = self.fwts.stdout.readline()
                percent = int(line.strip())
            except ValueError:
                continue

            msg = []
            line = self.fwts.stdout.readline()
            while line != 'XXX\n':
                if not line:
                    break

                msg.append(line)
                line = self.fwts.stdout.readline()

            msg = '<br/>'.join(msg)
            self.app_ctrl.status.progress(percent / 100.0, msg)

        del msg
        del line
        del percent

        result = json.load(open(path))

        summary = result['fwts'][-1]['summary']
        self.whiteboard = summary[0]['log_text']

        # Yeah, we only consider Critical and High failures as failure here
        # Ignored are Medium and Low failures
        fail_test = False
        del result
        for item in summary:
            if 'failure' in item:
                text = item['failure'][0]['log_text']

                if text.endswith('NONE'):
                    continue

                severity = text.split(' ', 1)[0]

                if severity in ['Critical', 'High', 'Medium']:
                    self.log.warn(text)
                else:
                    self.log.info(text)

                if len(item['failure']) > 1 and 'failures' in item['failure'][1]:
                    for failure in item['failure'][1]['failures']:
                        if severity in ['Critical', 'High', 'Medium']:
                            self.log.warn('    ' + failure['log_text'])
                            if not fail_test and severity in ['Critical']:

                                if 'temperature above threshold, cpu clock throttled' in text:
                                    # Do not test fail if CPU was throttled due to over-temperature
                                    pass
                                elif 'Critical failures:' in text:
                                    # Ignore the summary text
                                    pass
                                else:
                                    fail_test = text
                        else:
                            self.log.info('    ' + failure['log_text'])

        # The return code will pretty much always be a failure, so extract some
        # information from the log and warn/fail accordingly.
        if self.fwts.returncode != 0:
            self.log.warn(summary[0]['log_text'])
        else:
            self.log.info(summary[0]['log_text'])

        self.log.info('Please look at fwts.json for full test results.')

        del text
        del item
        del summary

        if fail_test:
            self.fail(fail_test)

if __name__ == "__main__":
    main()

