#!/usr/bin/env python2
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See LICENSE for more details.
#
# Copyright: Red Hat Inc. 2017
# Author: Benjamin Berg <bberg@redhat.com>

import os
import os.path
import re
import evdev
import select

from avocado import main
from avocado.core import exceptions

from fed_laptoptest.test import SessionTest, batch
from fed_laptoptest.hwinfo import HWinfo
from fed_laptoptest.ui import Signal, Event, TestUI

from gi.repository import Gtk, GLib

import time
import sys
import shutil

kbd_config_path = "/org/gnome/desktop/input-sources/sources"
kbd_config_value = [('xkb', 'us')]

class GnomeBatteryBench(SessionTest):
    """
    .. title:: Battery Runtime Test

    This test will run gnome-battery-bench to test the power usage of the
    laptop. The test is split into two parts (idle and light usage) and each
    part will run for about 10 minutes. The test requires the power to be
    unplugged as the battery discharge is used to estimate the power consumption.

    The keyboard layout will be set to US for the duration of the test.

    1. .. class:: action-wait-user

       Please unplug the power of the laptop to start the test.

    2. .. class:: action-test

       Please wait for the test to complete (no attention will be required for
       20 minutes).

    The test will finish automatically once it is done.

    :avocado: enable
    :avocado: tags=manual
    :param brightness: The brightness of the dispaly in percent (default: 50)
    :param duration: The duration of the test (default: 10min)
    :categories: battery
    """

    def setUp(self):
        if not os.path.exists('/usr/bin/gbb'):
            self.skip('GNOME Battery Bench is not installed, skipping test.')

        if not os.path.exists('/dev/uinput'):
            self.skip('Kernel needs to be compiled with uinput support, skipping test.')

        SessionTest.setUp(self)

        self.orig_kbd_config = self.session.get_dconf_key_raw(kbd_config_path)

    def tearDown(self):
        self.session.set_dconf_key_raw(kbd_config_path, self.orig_kbd_config)

        SessionTest.tearDown(self)

    def run_test(self, test):
        brightness = self.params.get('brightness', default=50)
        duration = self.params.get('duration', default='10min')

        self.session.set_dconf_key(kbd_config_path, kbd_config_value)

        self.app_ctrl.status.mark_action('wait-user')

        # Make sure that all the dbus services exist
        time.sleep(2)

        path = os.path.join(self.outputdir, 'gbb.json')
        proc, output = self.session.spawn_subprocess(
                lambda : os.execl('/usr/bin/gbb',
                                    'gbb', 'test', test,
                                    '-d', str(duration),
                                    '--screen-brightness', '50',
                                    '-o', '/tmp/gbb.json'))

        while proc.is_alive():
            r, w, x = self.app_ctrl.select([output])
            if output in r:
                line = os.read(output, 1000)
                sys.stderr.write(line)
                if line.startswith('Running;'):
                    self.app_ctrl.status.mark_action('test')

        if proc.exitcode != 0:
            self.fail('gbb process exited with error code %i' % proc.exitcode)
        shutil.move('/tmp/gbb.json', path)

    @batch
    def test_idle(self):
        self.run_test('idle')

    @batch
    def test_lightduty(self):
        self.run_test('lightduty')


if __name__ == "__main__":
    main()

