#!/usr/bin/env python2
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See LICENSE for more details.
#
# Copyright: Red Hat Inc. 2017
# Author: Benjamin Berg <bberg@redhat.com>

from avocado import main
from avocado.core import exceptions

from fed_laptoptest.test import SessionTest, interactive
from fed_laptoptest.ui import Signal, Event, SubtestSignal, SubtestDoneEvent, QuitSignal, TestUI
from fed_laptoptest.utils.touchpad import Touchpad

from gi.repository import Gtk, Gdk, GLib
import os
import os.path
import time
import subprocess

class TouchpadInfoSignal(Signal):
    def __init__(self, touchpad):
        # We must not send the touchpad over the wire as the other side will
        # try to close the FD (and will close something unrelated).
        self.fileno = touchpad.fileno()

        self.name = touchpad.name

        self.cap_abs = touchpad.cap_abs

        self.multitouch = touchpad.multitouch

        self.kernel_x = touchpad.kernel_x
        self.kernel_y = touchpad.kernel_y

        self.measured_x = touchpad.measured_x
        self.measured_y = touchpad.measured_y

class TouchpadSizeEvent(Event):
    def __init__(self, eventfd, width, height):
        self.eventfd = eventfd
        self.width = width
        self.height = height

dconf_path = '/org/gnome/desktop/peripherals/touchpad/'
# Older GNOME shell versions, the relevant keys are the same
dconf_path_legacy = '/org/gnome/settings-daemon/peripherals/touchpad/'

# Check if GTK+ is new enough for us
have_touchpad_gestures = hasattr(Gdk.EventMask, 'TOUCHPAD_GESTURE_MASK')

configs = {
    'normal' : {
      'send-events' : True,
      'left-handed' : False,
      'edge-scrolling-enabled' : False,
      'two-finger-scrolling-enabled' : True,
      'horiz-scroll-enabled' : True,
      'tap-to-click' : True,
      'tap-and-drag' : True,
      'natural-scroll' : True,
      'disable-while-typing' : True,
      # Legacy setting
      'scroll-method' : 'two-finger-scrolling',
    },
    'two-finger-non-natural' : {
      'two-finger-scrolling-enabled' : True,
      'edge-scrolling-enabled' : False,
      'natural-scroll' : False,
      # Legacy setting
      'scroll-method' : 'two-finger-scrolling',
    },
    'edge-natural' : {
      'two-finger-scrolling-enabled' : False,
      'edge-scrolling-enabled' : True,
      'natural-scroll' : True,
      # Legacy setting
      'scroll-method' : 'edge-scrolling',
    },
    'edge-non-natural' : {
      'two-finger-scrolling-enabled' : False,
      'edge-scrolling-enabled' : True,
      'natural-scroll' : False,
      # Legacy setting
      'scroll-method' : 'edge-scrolling',
    },
    'no-tap' : {
      'tap-to-click' : False,
    },
}

test_configs = ['normal', 'two-finger-non-natural', 'edge-natural', 'edge-non-natural', 'no-tap']

if set(test_configs) ^ set(configs.keys()):
    raise AssertionError('Have configurations that are not in test list.')


class GtkTouchpadUI(TestUI):

    def check_smooth_scroll(self, event, direction, sign=0):
        event = event.scroll
        if direction == 'horizontal' and abs(event.delta_x) > abs(event.delta_y):
            if sign >= 0 and event.delta_x > 0:
                return True
            if sign <= 0 and event.delta_x < 0:
                return True

        if direction == 'vertical' and abs(event.delta_y) > abs(event.delta_x):
            if sign >= 0 and event.delta_y > 0:
                return True
            if sign <= 0 and event.delta_y < 0:
                return True

        return False

    def check_motion(self, event, direction, sign=0):
        event = event.motion

        # This will be called multiple times for one event
        if not hasattr(self, 'last_motion') or event.time - self.last_pos[2] > 100:
            self.last_motion = (event.x, event.y, event.time)
            self.last_pos = self.last_motion
            return

        new_motion = (event.x, event.y, event.time)
        if self.last_motion != new_motion:
            self.last_pos = self.last_motion

        d_x = event.x - self.last_pos[0]
        d_y = event.y - self.last_pos[1]
        self.last_motion = new_motion

        if direction == 'horizontal' and abs(d_x) > abs(d_y):
            if sign >= 0 and d_x > 0:
                return True
            if sign <= 0 and d_x < 0:
                return True

        if direction == 'vertical' and abs(d_y) > abs(d_x):
            if sign >= 0 and d_y > 0:
                return True
            if sign <= 0 and d_y < 0:
                return True

        return False

    last_motion_time = 0
    pressed_time = 0
    pressed = set()
    press_and_motion = False

    def check_motion_blocked(self, event, detect_failure):
        detected_block = False
        detected_failure = False

        if event.type == Gdk.EventType.MOTION_NOTIFY:
            self.last_motion_time = event.motion.time

        if event.type == Gdk.EventType.KEY_PRESS:
            if event.keyval not in self.pressed:
                # If the time is the same, then it is the repeated event
                if not self.pressed:
                    self.pressed_time = event.key.time
                self.pressed.add(event.keyval)
                if self.pressed_time - self.last_motion_time < 50:
                    self.press_and_motion = True
                else:
                    self.press_and_motion = False

        if event.type == Gdk.EventType.KEY_RELEASE:
            # https://bugzilla.gnome.org/show_bug.cgi?id=784611
            if event.keyval in self.pressed:
                self.pressed.remove(event.keyval)
            if not self.pressed:
                if self.last_motion_time > self.pressed_time:
                    detected_failure = True
                elif self.press_and_motion:
                    self.press_and_motion = False
                    detected_block = True

                self.press_and_motion = False

        if detect_failure:
            return detected_failure
        return detected_block


    expected_events = [
        # Motion event
        {
            'desc' : 'Motion from left to right',
            'type' : Gdk.EventType.MOTION_NOTIFY,
            'evt' : 'motion',
            'func' : check_motion,
            'func_args' : ('horizontal', 1),
            'config' : 'normal'
        },
        {
            'desc' : 'Motion from right to left',
            'type' : Gdk.EventType.MOTION_NOTIFY,
            'evt' : 'motion',
            'func' : check_motion,
            'func_args' : ('horizontal', -1),
            'config' : 'normal'
        },
        {
            'desc' : 'Motion from top to bottom',
            'type' : Gdk.EventType.MOTION_NOTIFY,
            'evt' : 'motion',
            'func' : check_motion,
            'func_args' : ('vertical', 1),
            'config' : 'normal'
        },
        {
            'desc' : 'Motion from bottom to top',
            'type' : Gdk.EventType.MOTION_NOTIFY,
            'evt' : 'motion',
            'func' : check_motion,
            'func_args' : ('vertical', -1),
            'config' : 'normal'
        },

        # Tap
        {
            'desc' : 'Tap left button (one finger) to click',
            'type' : Gdk.EventType.BUTTON_PRESS,
            'evt' : 'button',
            'button' : 1,
            'config' : 'normal',
        },
        {
            'desc' : 'Tap right button (two fingers) to click',
            'type' : Gdk.EventType.BUTTON_PRESS,
            'evt' : 'button',
            'button' : 3,
            'config' : 'normal',
        },
        {
            'desc' : 'Tap middle button (three fingers) to click',
            'type' : Gdk.EventType.BUTTON_PRESS,
            'evt' : 'button',
            'button' : 2,
            'config' : 'normal',
        },

        # Normal button click
        {
            'desc' : 'Left hardware button click',
            'type' : Gdk.EventType.BUTTON_PRESS,
            'evt' : 'button',
            'button' : 1,
            'config' : 'no-tap',
            'optional' : True
        },
        {
            'desc' : 'Right hardware button click',
            'type' : Gdk.EventType.BUTTON_PRESS,
            'evt' : 'button',
            'button' : 3,
            'config' : 'no-tap',
            'optional' : True
        },
        {
            'desc' : 'Middle hardware button click',
            'type' : Gdk.EventType.BUTTON_PRESS,
            'evt' : 'button',
            'button' : 2,
            'config' : 'no-tap',
            'optional' : True
        },

        # Click + drag (only left click)
        {
            'desc' : 'Tap left button (one finger) to click and drag',
            'type' : Gdk.EventType.MOTION_NOTIFY,
            'evt' : 'motion',
            'state_mask' : Gdk.ModifierType.BUTTON1_MASK,
            'config' : 'normal',
        },

        # Scroll usually reversed (natural scrolling)
        {
            'desc' : 'Two finger scroll, fingers moving left',
            'type' : Gdk.EventType.SCROLL,
            'evt' : 'scroll',
            'direction' : Gdk.ScrollDirection.SMOOTH,
            'func' : check_smooth_scroll,
            'func_args' : ('horizontal', 1),
            'config' : 'normal'
        },
        {
            'desc' : 'Two finger scroll, fingers moving right',
            'type' : Gdk.EventType.SCROLL,
            'evt' : 'scroll',
            'direction' : Gdk.ScrollDirection.SMOOTH,
            'func' : check_smooth_scroll,
            'func_args' : ('horizontal', -1),
            'config' : 'normal'
        },
        {
            'desc' : 'Two finger scroll, fingers moving down',
            'type' : Gdk.EventType.SCROLL,
            'evt' : 'scroll',
            'direction' : Gdk.ScrollDirection.SMOOTH,
            'func' : check_smooth_scroll,
            'func_args' : ('vertical', -1),
            'config' : 'normal'
        },
        {
            'desc' : 'Two finger scroll, fingers moving up',
            'type' : Gdk.EventType.SCROLL,
            'evt' : 'scroll',
            'direction' : Gdk.ScrollDirection.SMOOTH,
            'func' : check_smooth_scroll,
            'func_args' : ('vertical', 1),
            'config' : 'normal'
        },

        # Scroll not reversed if natural scrolling is disabled
        {
            'desc' : 'Two finger scroll, fingers moving left',
            'type' : Gdk.EventType.SCROLL,
            'evt' : 'scroll',
            'direction' : Gdk.ScrollDirection.SMOOTH,
            'func' : check_smooth_scroll,
            'func_args' : ('horizontal', -1),
            'config' : 'two-finger-non-natural'
        },
        {
            'desc' : 'Two finger scroll, fingers moving right',
            'type' : Gdk.EventType.SCROLL,
            'evt' : 'scroll',
            'direction' : Gdk.ScrollDirection.SMOOTH,
            'func' : check_smooth_scroll,
            'func_args' : ('horizontal', 1),
            'config' : 'two-finger-non-natural'
        },
        {
            'desc' : 'Two finger scroll, fingers moving down',
            'type' : Gdk.EventType.SCROLL,
            'evt' : 'scroll',
            'direction' : Gdk.ScrollDirection.SMOOTH,
            'func' : check_smooth_scroll,
            'func_args' : ('vertical', 1),
            'config' : 'two-finger-non-natural'
        },
        {
            'desc' : 'Two finger scroll, fingers moving up',
            'type' : Gdk.EventType.SCROLL,
            'evt' : 'scroll',
            'direction' : Gdk.ScrollDirection.SMOOTH,
            'func' : check_smooth_scroll,
            'func_args' : ('vertical', -1),
            'config' : 'two-finger-non-natural'
        },

        # Edge scroll with natural scrolling enabled
        {
            'desc' : 'Lower edge scroll, finger moving left',
            'type' : Gdk.EventType.SCROLL,
            'evt' : 'scroll',
            'direction' : Gdk.ScrollDirection.SMOOTH,
            'func' : check_smooth_scroll,
            'func_args' : ('horizontal', 1),
            'config' : 'edge-natural'
        },
        {
            'desc' : 'Lower edge scroll, finger moving right',
            'type' : Gdk.EventType.SCROLL,
            'evt' : 'scroll',
            'direction' : Gdk.ScrollDirection.SMOOTH,
            'func' : check_smooth_scroll,
            'func_args' : ('horizontal', -1),
            'config' : 'edge-natural'
        },
        {
            'desc' : 'Right edge scroll, finger moving down',
            'type' : Gdk.EventType.SCROLL,
            'evt' : 'scroll',
            'direction' : Gdk.ScrollDirection.SMOOTH,
            'func' : check_smooth_scroll,
            'func_args' : ('vertical', -1),
            'config' : 'edge-natural'
        },
        {
            'desc' : 'Right edge scroll, finger moving up',
            'type' : Gdk.EventType.SCROLL,
            'evt' : 'scroll',
            'direction' : Gdk.ScrollDirection.SMOOTH,
            'func' : check_smooth_scroll,
            'func_args' : ('vertical', 1),
            'config' : 'edge-natural'
        },

        # Edge scroll with natural scrolling disabled
        {
            'desc' : 'Lower edge scroll, finger moving left',
            'type' : Gdk.EventType.SCROLL,
            'evt' : 'scroll',
            'direction' : Gdk.ScrollDirection.SMOOTH,
            'func' : check_smooth_scroll,
            'func_args' : ('horizontal', -1),
            'config' : 'edge-non-natural'
        },
        {
            'desc' : 'Lower edge scroll, finger moving right',
            'type' : Gdk.EventType.SCROLL,
            'evt' : 'scroll',
            'direction' : Gdk.ScrollDirection.SMOOTH,
            'func' : check_smooth_scroll,
            'func_args' : ('horizontal', 1),
            'config' : 'edge-non-natural'
        },
        {
            'desc' : 'Right edge scroll, finger moving down',
            'type' : Gdk.EventType.SCROLL,
            'evt' : 'scroll',
            'direction' : Gdk.ScrollDirection.SMOOTH,
            'func' : check_smooth_scroll,
            'func_args' : ('vertical', 1),
            'config' : 'edge-non-natural'
        },
        {
            'desc' : 'Right edge scroll, finger moving up',
            'type' : Gdk.EventType.SCROLL,
            'evt' : 'scroll',
            'direction' : Gdk.ScrollDirection.SMOOTH,
            'func' : check_smooth_scroll,
            'func_args' : ('vertical', -1),
            'config' : 'edge-non-natural'
        },

        # Keyboard interaction
        {
            'desc' : 'Press and release a character on internal keyboard while moving finger over touchpad (should block motion)',
            'func' : check_motion_blocked,
            'func_args' : (False,),
            'config' : 'normal'
        },
    ]

    if have_touchpad_gestures:
        expected_events.extend([
            # Pinch/Swipe actions
            {
                'desc' : 'Pinch Gesture (two finger zoom)',
                'type' : Gdk.EventType.TOUCHPAD_PINCH,
                'evt' : 'touchpad_pinch',
                'config' : 'normal'
            },

            {
                'desc' : 'Swipe Gesture (three or more fingers)',
                'type' : Gdk.EventType.TOUCHPAD_SWIPE,
                'evt' : 'touchpad_swipe',
                # We don't seem to get proper finger counts often
                'config' : 'normal'
            },
        ])


    def __init__(self, app):
        self.app = app
        self.current_config = None

        self._marked_events = dict()
        self._detected_events = list()

    def unmark_event(self, ee):
        del self._marked_events[ee['desc']]

        self.update_label()

    def mark_event(self, ee):
        was_marked = False
        if ee['desc'] in self._marked_events:
            was_marked = True
            GLib.source_remove(self._marked_events[ee['desc']])

        self._marked_events[ee['desc']] = GLib.timeout_add(500, self.unmark_event, ee)
        if ee not in self._detected_events:
            self._detected_events.append(ee)
        self._detected_events.sort(key=lambda x : x['desc'])

        self.update_label()

    def event_cb(self, widget, event):
        if event.type == Gdk.EventType.KEY_PRESS:
            if event.keyval == Gdk.KEY_Escape:
                self.event_box.grab_remove()
                self.send_subtest_status()
                self.app.send(SubtestDoneEvent(self.current_config, prompt=True))
            if event.keyval in [Gdk.KEY_Return, Gdk.KEY_KP_Enter]:
                self.event_box.grab_remove()
                failure = self.send_subtest_status()
                self.app.send(SubtestDoneEvent(self.current_config, failure=failure))


        for ee in self.expected_events:
            if 'config' in ee and ee['config'] != self.current_config:
                continue

            if 'evt' in ee:
                evt = getattr(event, ee['evt'])
            else:
                evt = event

            for k, v in ee.iteritems():
                if k in ['desc', 'evt', 'func', 'func_args', 'config', 'optional']:
                    continue

                if k.endswith('_mask'):
                    if not hasattr(evt, k[:-5]):
                        break
                    val = getattr(evt, k[:-5])
                    if val & v != v:
                        break
                else:
                    if not hasattr(evt, k):
                        break
                    val = getattr(evt, k)
                    if val != v:
                        break

            else:
                # No break happened, i.e. event is good.
                match = True
                if 'func' in ee:
                    match = ee['func'](self, event, *ee['func_args'])
                if match:
                    self.mark_event(ee)

        # Do not pass events on (otherwise accidental text will be entered or
        # buttons pressed). Ungrab happens when required.
        return True

    def map_event_cb(self, widget, event):
        self.event_box.get_window().set_event_compression(False)
        self.event_box.grab_focus()
        #self.event_box.grab_add()

    def send_subtest_status(self):
        failure = False
        for ee in self.expected_events:
            if ee['config'] != self.current_config:
                continue

            if ee in self._detected_events:
                self.app.info('Detected: %s' % ee['desc'])
            else:
                if not 'optional' in ee or not ee['optional']:
                    failure = True
                    self.app.warning('Not detected: %s' % ee['desc'])
                else:
                    self.app.info('Not detected: %s' % ee['desc'])

        return failure

    def update_label(self):
        if self.current_config is None:
            self.info_label.set_markup('')
            return

        items = []

        for ee in self.expected_events:
            if ee['config'] != self.current_config:
                continue
            tags=[]
            if ee in self._detected_events:
                tags.append('fgcolor="black"')
                tags.append('bgcolor="lightgreen"')
            elif 'optional' in ee and ee['optional']:
                tags.append('fgcolor="black"')
                tags.append('bgcolor="lightyellow"')
            if ee['desc'] in self._marked_events:
                tags.append('weight="bold"')

            items.append(' * <span %s>%s</span>' % (' '.join(tags), ee['desc']))
        text = '<span size="larger">Please do the following actions:</span>\n'
        text += '\n'.join(items)
        self.info_label.set_markup(text)

    def get_widget(self):
        vbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)

        label = Gtk.Label(xalign=0, yalign=0)
        label.set_markup('<b>Detected Touchpads</b>')
        vbox.add(label)

        scroll = Gtk.ScrolledWindow()
        scroll.set_size_request(-1, 200)
        self.listbox = Gtk.ListBox()
        scroll.add(self.listbox)
        vbox.add(scroll)

        self.info_label = Gtk.Label(xalign=0, yalign=0)
        self.update_label()

        self.event_box = Gtk.EventBox(above_child=True)
        self.event_box.props.can_focus = True
        if have_touchpad_gestures:
            self.event_box.add_events(Gdk.EventMask.TOUCHPAD_GESTURE_MASK)
        self.event_box.add_events(
            Gdk.EventMask.POINTER_MOTION_MASK |
            Gdk.EventMask.BUTTON_MOTION_MASK |
            Gdk.EventMask.BUTTON_PRESS_MASK |
            Gdk.EventMask.BUTTON_RELEASE_MASK |
            Gdk.EventMask.KEY_PRESS_MASK |
            Gdk.EventMask.KEY_RELEASE_MASK |
            Gdk.EventMask.SMOOTH_SCROLL_MASK
        )

        self.event_box.connect('map-event', self.map_event_cb)
        self.event_box.connect('event', self.event_cb)

        vbox.pack_start(self.event_box, True, True, 0)

        self.event_box.add(self.info_label)
        self.event_box.show_all()

        return vbox

    def apply_config(self):
        import subprocess

        for path in [dconf_path, dconf_path_legacy]:
            if self.current_config is None:
                subprocess.Popen(['dconf', 'reset', '-f', path])
            else:
                for k, v in configs[self.current_config].iteritems():
                    if isinstance(v, bool):
                        subprocess.Popen(['dconf', 'write', path+k, 'true' if v else 'false'])
                    else:
                        subprocess.Popen(['dconf', 'write', path+k, "'" + v + "'"])

    def recv(self, event):
        if isinstance(event, QuitSignal):
            self.current_config = None
            self.apply_config()

        if isinstance(event, SubtestSignal):
            self.current_config = event.subtest

            for k, v in self._marked_events.iteritems():
                GLib.source_remove(v)

            self._marked_events = dict()
            self._detected_events = list()

            self.apply_config()

            self.update_label()

            self.event_box.grab_add()

        if isinstance(event, TouchpadInfoSignal):
            for row in self.listbox.get_children():
                if row.touchpad.fileno == event.fileno:
                    break
            else:
                row = Gtk.ListBoxRow()
                self.listbox.add(row)

            row.touchpad = event
            self.update_touchpad_row(row)

    def touchpad_size_changed_cb(self, spin, row):
        self.app.send(TouchpadSizeEvent(row.touchpad.fileno,
                                        row.spin_width.get_value(),
                                        row.spin_height.get_value()))

    def update_touchpad_row(self, row):
        vbox = row.get_child()
        if not vbox:
            vbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
            vbox.set_spacing(5)
            row.add(vbox)

            row.name = Gtk.Label()
            vbox.add(row.name)
            row.name.set_markup('<b>{:s}</b>'.format(GLib.markup_escape_text(row.touchpad.name)))
            row.name.set_halign(Gtk.Align.START)

            row.detected = Gtk.Label()
            row.detected.set_margin_start(5)
            row.detected.set_hexpand(True)
            row.detected.set_halign(Gtk.Align.START)
            row.detected.set_text('Detected horizontal range is from {:d} to {:d} (resolution: {:d}, width {:.1f}mm)\n'
                                  'Detected vertical range is from {:d} to {:d} (resolution: {:d}, height {:.1f}mm)'.format(
                                      row.touchpad.kernel_x.min, row.touchpad.kernel_x.max,
                                      row.touchpad.kernel_x.resolution, row.touchpad.kernel_x.size,
                                      row.touchpad.kernel_y.min, row.touchpad.kernel_y.max,
                                      row.touchpad.kernel_y.resolution, row.touchpad.kernel_y.size,
                                   ))
            vbox.add(row.detected)

            row.measured = Gtk.Label()
            row.measured.set_margin_start(5)
            row.measured.set_halign(Gtk.Align.START)
            vbox.add(row.measured)

            hbox = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)
            vbox.add(hbox)
            hbox.set_spacing(1)
            hbox.set_margin_start(5)
            hbox.set_halign(Gtk.Align.START)

            hbox.add(Gtk.Label('Please measure the width '))
            row.spin_width = Gtk.SpinButton()
            row.spin_width.set_range(0, 300)
            row.spin_width.set_digits(1)
            row.spin_width.set_increments(1, 10)
            row.spin_width.connect('changed', self.touchpad_size_changed_cb, row)
            hbox.add(row.spin_width)

            hbox.add(Gtk.Label('mm and height '))
            row.spin_height = Gtk.SpinButton(climb_rate=1, digits=1)
            row.spin_height.set_range(0, 300)
            row.spin_height.set_digits(1)
            row.spin_height.set_increments(1, 10)
            row.spin_height.connect('changed', self.touchpad_size_changed_cb, row)
            hbox.add(row.spin_height)
            hbox.add(Gtk.Label('mm.'))

        # This part is dynamic
        if row.touchpad.measured_x.min < row.touchpad.measured_x.max and \
           row.touchpad.measured_y.min < row.touchpad.measured_y.max:
            row.measured.set_text('Measured horizontal range is from {:d} to {:d} (resolution: {:d}, width {:.1f}mm)\n'
                                  'Measured vertical range is from {:d} to {:d} (resolution: {:d}, height {:.1f}mm)'.format(
                                      row.touchpad.measured_x.min, row.touchpad.measured_x.max,
                                      row.touchpad.measured_x.resolution, row.touchpad.measured_x.size,
                                      row.touchpad.measured_y.min, row.touchpad.measured_y.max,
                                      row.touchpad.measured_y.resolution, row.touchpad.measured_y.size,
                                   ))
        else:
            row.measured.set_text('Waiting for events.')

        row.show_all()

class GtkTouchpad(SessionTest):
    """
    .. title:: Touchpad (GTK+ based)

    In this test we check that the touchpad generated events are visible by
    GTK+ and that the GNOME settings can be modified to modify the touchpads
    behaviour.

    Please note that the test only shows whether certain events are detected.
    It cannot tell a normal mouse movement from a touchpad movement, so be
    careful to understand that misdetections can happen and you need to
    manually fail or warn with an explanation if this is the case.

    .. class:: important

        During each phase of the test press **Enter** to confirm success (or
        failure). Press **Esc** to enter additional information.

    .. class:: action-size

       All touchpads are listed on the right side. Please check that your
       touchpad has been detected. If not, please *Fail* the test. If you do
       not have any touchpad then the test will finish.

       Please run your finger along the edge of your touchpad a few times. Doing
       so the minimum and maximum reported values are measured. Please also
       measure the size in mm and enter it into the provided fields.

    .. class:: action-normal

        Tests running under a usual configuration. This means:

        * Tap to click and tap to drag is enabled
        * Natural scrolling is enabled
        * Two finger scrolling is enabled
        * Disable while typing is enabled

    .. class:: action-two-finger-non-natural

        Confirm that natural scrolling can be disabled

    .. class:: action-edge-natural

        Confirm that edge scrolling works correctly with natural scrolling enabled.

    .. class:: action-edge-non-natural

        Confirm that edge scrolling works correctly with natural scrolling disabled.

    .. class:: action-no-tap

        Confirm that tap to click is disabled. Please confirm that tap to click
        does *not* work and then click the hardware buttons.

    .. class:: action-finish

       Please confirm the test results.

    :avocado: enable
    :avocado: tags=manual
    :categories: pointer
    """

    ui_class = GtkTouchpadUI

    def setUp(self):
        SessionTest.setUp(self)

        # asdf
        self.touchpads = {}
        for touchpad in Touchpad.find_touchpads():
            self.touchpads[touchpad.fileno()] = touchpad
            self.app_ctrl.send(TouchpadInfoSignal(touchpad))

        self.evemu_procs = []

    def tearDown(self):
        SessionTest.tearDown(self)

        for proc in self.evemu_procs:
            proc.terminate()
        time.sleep(0.1)
        for proc in self.evemu_procs:
            if proc.poll() is None:
                proc.kill()

        for touchpad in self.touchpads.itervalues():
            self.log.info(str(touchpad))
            self.log.info('Reported X range for Touchpad from {:d} to {:d} (resolution {:d}, size {:.4f}mm).'.format(touchpad.kernel_x.min, touchpad.kernel_x.max, touchpad.kernel_x.resolution, touchpad.kernel_x.size))
            self.log.info('User X range for Touchpad from {:d} to {:d} (resolution {:d}, size {:.4f}mm).'.format(touchpad.measured_x.min, touchpad.measured_x.max, touchpad.measured_x.resolution, touchpad.measured_x.size))
            self.log.info('Reported Y range for Touchpad from {:d} to {:d} (resolution {:d}, size {:.4f}mm).'.format(touchpad.kernel_y.min, touchpad.kernel_y.max, touchpad.kernel_y.resolution, touchpad.kernel_y.size))
            self.log.info('User Y range for Touchpad from {:d} to {:d} (resolution {:d}, size {:.4f}mm).'.format(touchpad.measured_y.min, touchpad.measured_y.max, touchpad.measured_y.resolution, touchpad.measured_y.size))

    def recv(self, event):
        if isinstance(event, TouchpadSizeEvent):
            touchpad = self.touchpads[event.eventfd]
            touchpad.set_size(event.width, event.height)

            self.app_ctrl.send(TouchpadInfoSignal(touchpad))

        elif isinstance(event, SubtestDoneEvent):
            self.subtest_done = True

            if 'prompt' in event.result:
                self.app_ctrl.status.wait_continue()
            if 'failure' in event.result and event.result['failure']:
                self.fail('Subtest {:s} failed'.format(self.subtest))

    @interactive
    def test(self):
        global test_configs
        failures = []

        for touchpad in self.touchpads.itervalues():
            devname = os.path.basename(touchpad.fn)
            outprefix = os.path.join(self.outputdir, devname)
            try:
                proc = subprocess.Popen(['evemu-record', '--autorestart=3', touchpad.fn, outprefix])
            except OSError:
                pass
            self.evemu_procs.append(proc)

        self.app_ctrl.status.mark_action('size')
        self.app_ctrl.status.allow_continue()

        while not self.app_ctrl.status.continued:
            r, w, x = self.app_ctrl.select(self.touchpads.keys())
            for fd in r:
                self.touchpads[fd].read_one()
                self.app_ctrl.send(TouchpadInfoSignal(self.touchpads[fd]))

        # User hit continue and no touchpads were detected
        if not self.touchpads:
            self.skip_user_confirm = True
            return

        self.proceed = False
        for test in test_configs:
            self.app_ctrl.status.mark_action(test)
            self.app_ctrl.send(SubtestSignal(test))
            self.subtest_done = False
            self.subtest = test

            while not self.subtest_done:
                try:
                    r, w, x = self.app_ctrl.select(self.touchpads.keys())
                    for fd in r:
                        self.touchpads[fd].read_one()
                        self.app_ctrl.send(TouchpadInfoSignal(self.touchpads[fd]))

                except exceptions.TestFail as failure:
                    self.log.warning('Got test failure for subtest %s: %s' % (test, str(failure)))
                    self.subtest_done = True
                    failures.append(failure)

        if failures:
            raise failures[0]

