#!/usr/bin/env python2
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See LICENSE for more details.
#
# Copyright: Red Hat Inc. 2017
# Author: Benjamin Berg <bberg@redhat.com>

import os
import os.path
import re
import evdev
import select
import time

from avocado import main
from avocado.core import exceptions

from fed_laptoptest.test import SessionTest, interactive
from fed_laptoptest.hwinfo import HWinfo
from fed_laptoptest.ui import Signal, Event, TestUI
from fed_laptoptest.utils import usb, ui as uiutils

from gi.repository import Gtk, GLib, NM, Gio

class SimpleUSBPortUI(TestUI):

    def update_usb(self, slot, it, parent_it):
        slot.stale = False

        if it is None:
            it = self.treestore.append(parent_it)
            obj = None
        else:
            obj = self.treestore[it][0]
            assert obj == slot

        if not isinstance(slot, usb.USBDevice) and isinstance(obj, usb.USBDevice):
            slot = obj

            if not slot.stale:
                self.app.info('Removed device at {:s}: {:s}'.format(list(slot.paths())[-1], str(obj)))

            slot.stale = True
        elif obj is None or (isinstance(slot, usb.USBDevice) and not isinstance(obj, usb.USBDevice)):
            self.app.info('New entry at {:s}: {:s}'.format(list(slot.paths())[-1], str(slot)))

        markup = GLib.markup_escape_text(str(slot))
        if slot.stale:
            markup = '<span style="italic">{:s}</span>'.format(markup)
        self.treestore.set_row(it, [slot, markup])

        if slot.stale or not isinstance(slot, usb.USBHub):
            # Remove all children (i.e. in case a hub was unplugged)
            child_it = self.treestore.iter_children(it)
            while child_it:
                if not self.treestore.remove(child_it):
                    child_it = None

            return

        child_it = self.treestore.iter_children(it)
        for obj in slot.ports:
            self.update_usb(obj, child_it, it)
            if child_it:
                child_it = self.treestore.iter_next(child_it)

        # Remove if there were more slots than there are now (should never
        # happen)
        while child_it:
            child_it = self.treestore.remove(child_it)

    def update_usb_tree(self):
        roots = list(usb.get_roots())

        it = self.treestore.iter_children(None)
        while it:
            obj = self.treestore[it][0]
            for root in roots:
                if root == obj:
                    self.update_usb(root, it, None)
                    roots.remove(root)
                    it = self.treestore.iter_next(it)
                    break
            else:
                if not self.treestore.remove(it):
                    it = None

        # Anything not handled yet needs to be added
        for root in roots:
            self.update_usb(root, None, None)

        self.treeview.expand_all()

        return GLib.SOURCE_CONTINUE

    def get_widget(self):
        scroll = Gtk.ScrolledWindow()
        self.treeview = Gtk.TreeView()
        self.treestore = Gtk.TreeStore(object, str)

        self.treeview.set_model(self.treestore)

        renderer = Gtk.CellRendererText()
        col = Gtk.TreeViewColumn("USB Device", renderer, markup=1)
        self.treeview.append_column(col)

        scroll.add(self.treeview)
        scroll.show_all()

        GLib.timeout_add(1000, self.update_usb_tree)

        return scroll



class SimpleUSBPortTest(SessionTest):
    """
    .. title:: USB Port Test

    This is a simple test to verify that all the USB ports work. If you attach
    a device to a port then it will show and the information about it will be
    retained after unplugging.

    .. note::

       It is entirely normal if there are ports listed which are not connected
       to anything. USB 2 ports may also be listed as supporting USB 3.

    .. class:: action-finish

       If you can, plug in a device into each port of your laptop. Feel free
       to simply *Cancel* or *Finish* the test. This test is entirely manual.

    :avocado: enable
    :avocado: tags=manual,userfacing
    """

    ui_class = SimpleUSBPortUI

    @interactive
    def test(self):
        self.app_ctrl.status.mark_action('finish')

        self.app_ctrl.status.wait_finished()


if __name__ == "__main__":
    main()

